/* HPCU Bi-Weekly Challenge: OpenCL Pi
 * Author: Aaron Weeden, Shodor, April 2015
 *         Based on https://shodor.org/media/content//petascale/materials/BW2014/EveningLabs/pi_c.c
 * Description:
 * This program is an exercise in approximating pi.
 *
 * The program uses an area-of-a-circle formula (A = pi * r^2) and a circle
 * equation (x^2 + y^2 = r^2) as well as a sum of areas of rectangles
 * (w1*h1 + w2*h2 + ...).
 *
 * The circle used has a radius of 1, so this simplifies to A = pi and
 * x^2 + y^2 = 1.
 * Also, each rectangle has the same width, so the sum of areas simplifies to
 * w*h1 + w*h2 + w*h3 + ...
 *
 * Furthermore, the rectangle heights are determined by their intersections with
 * the circle, so the heights are h1 = y1 = sqrt(1 - x1^2),
 * h2 = y2 = sqrt(1 - x2^2), ...
 *
 * Lastly, the total number of rectangles is simplified by using a unit circle
 * and focusing on the first quadrant. This divides the area by 4, so
 * pi = 4 * A.
 *
 * To summarize, the sum of the area of the rectangles, obtained as described,
 * provides an approximation for pi when multiplied by 4.
 */

/* LIBRARIES */
#include <stdio.h>  /* printf function */
#include <math.h>   /* sqrt function */
#include <float.h>  /* FLT_EPSILON and LDBL_DIG constants */

/* MAIN FUNCTION */
int main(int argc, char **argv) {
  int n_rects = 256; /* number of rectangles */
  float rect_width = 1.0 / n_rects; /* width of each rectangle */
  float rect_left = 0.0; /* left-most x-coordinate of each rectangle */
  float rect_height_squared; /* squared height of each rectangle */
  float rect_height; /* height of each rectangle */
  float agg_area = 0.0; /* total area of rectangles */
  float pi; /* final calculated value of pi */
  int i; /* for looping over rectangles */

  /* Loop over each of the rectangles */
  for (i = 0; i < n_rects; i++) {

    /* Advance the left-most coordinate by one width of a rectangle */
    rect_left += rect_width;

    /* y^2 = 1 - x^2, where y is the rectangle's height */
    rect_height_squared = 1 - rect_left*rect_left;

    /* 1 - x^2 may produce a value that is 0 but on a computer may come out as
     * almost, but not quite, zero. Make sure we use 0 moving forward.
     */
    if (rect_height_squared < FLT_EPSILON) {
      rect_height_squared = 0;
    }

    /* height = the square root of height^2 */
    rect_height = sqrt(rect_height_squared);

    /* Calculate the area of the rectangle and add it to the aggregate area */
    agg_area += rect_width * rect_height;

    /* Done looping over all the rectangles */
  }

  /* Calculate the final value of pi */
  pi = 4 * agg_area;

  /* Print the final value of pi */
  printf("%.*f\n", LDBL_DIG, pi);

  /* Exit the program */
  return 0;
}
